// Generated by CoffeeScript 2.3.2
(function() {
  var debug, pkg,
    hasProp = {}.hasOwnProperty;

  pkg = require('../../package.json');

  this.name = `${pkg.name}:middleware:carrier:setup`;

  debug = (require('tangible'))(this.name);

  // * doc.global_number Record with an identifier `number:<global-number>`. These records are used between the carrier SBCs and the client SBCs. They are one of the two types of `doc.number`.
  // * doc.number If the identifier of a number does not contain a `@` character, it is a `doc.global_number`.
  // * doc.global_number._id (required) `number:<global-number>`
  // * doc.global_number.number (required) `<global-number>`, where the global-number has the standard format `<country-code><national-number>`. The format is identical to the E.164 format but the `+` sign at the start is omitted.

  // Config
  // ======

  // Note that carrier-side the fields are called `sip_profiles` and are stored in the database.
  // The FreeSwitch configuration uses the `profiles` field, which defaults to using port 5080.
  this.config = async function() {
    var base, k, name, p, profile, ref, ref1, ref2, ref3, ref4, v;
    // Create the proper profiles and ACLs
    this.cfg.profiles = {};
    this.cfg.acls = {};
    this.cfg.host_data = this.cfg.host != null ? (debug(`Retrieving data for ${this.cfg.host}`), (await this.cfg.prov.get(`host:${this.cfg.host}`).catch((error) => {
      debug.ops(`Host ${this.cfg.host}: ${error}`);
      return {};
    }))) : (debug.dev('No cfg.host'), {});
    if ((base = this.cfg).sip_profiles == null) {
      base.sip_profiles = (ref = this.cfg.host_data.sip_profiles) != null ? ref : {};
    }
    debug('Configuring SIP Profiles', this.cfg.sip_profiles);
    ref1 = this.cfg.sip_profiles;
    for (name in ref1) {
      if (!hasProp.call(ref1, name)) continue;
      profile = ref1[name];
      p = {
        local_ip: profile.ingress_sip_ip,
        socket_port: (ref2 = this.cfg.port) != null ? ref2 : 5702
      };
      for (k in profile) {
        if (!hasProp.call(profile, k)) continue;
        v = profile[k];
        p[k] = v;
      }
      this.cfg.profiles[name] = p;
      this.cfg.acls[`${name}-ingress`] = (ref3 = profile.ingress_acl) != null ? ref3 : [];
      this.cfg.acls[`${name}-egress`] = (ref4 = profile.egress_acl) != null ? ref4 : [];
    }
    return null;
  };

  // Server
  // ======

  // Load the host record so that we can retrieve the `sip_profiles` at runtime.
  this.server_pre = async function() {
    var base, ref;
    this.cfg.host_data = this.cfg.host != null ? (debug(`Retrieving data for ${this.cfg.host}`), (await this.cfg.prov.get(`host:${this.cfg.host}`).catch((error) => {
      // debug.ops "Host #{@cfg.host}: #{error}"
      debug(`Host ${this.cfg.host}: ${error}`);
      return {};
    // debug.dev 'No cfg.host'
    }))) : (debug('No cfg.host'), {});
    if ((base = this.cfg).sip_profiles == null) {
      base.sip_profiles = (ref = this.cfg.host_data.sip_profiles) != null ? ref : {};
    }
    debug('Configuring SIP Profiles', this.cfg.sip_profiles);
    return null;
  };

  // Call-Processing
  // ===============
  this.include = async function() {
    var C, Reference, base, base1, base2, base3, base4, our_dialplan, ref, ref1, reference_in, sip_params, sip_profile_carrier, sip_profile_client;
    // Session Context
    // ---------------

    // * session.context (string) The original Sofia Context for this (ingress) call.

    // The channel-context is set (for calls originating from sofia-sip) by the `context` parameter of the Sofia instance that carries the A leg.
    if ((base = this.session).context == null) {
      base.context = this.data['Channel-Context'];
    }
    debug('>>>> New call', this.session.context, this.data);
    if (!(C = (ref = this.session.context) != null ? ref.match(/^(\S+)-(ingress|egress|transfer|handled)(?:-(\S+))?$/) : void 0)) {
      debug.dev('Malformed context', this.session.context);
      await this.respond('500 Malformed context');
      return;
    }
    this.session.profile = C[1];
    if ((base1 = this.session).reference == null) {
      base1.reference = C[3];
    }
    // Session Reference
    // -----------------

    // * session.reference (string) Identifies a call spanning multiple FreeSwitch servers.

    // The `reference` is used to track a given call through various systems and associate parameters (e.g. client information) to the call as a whole.
    // In case of a transfer, the session identifier might be included in the context.
    if ((base2 = this.session).reference == null) {
      base2.reference = this.req.variable('session_reference');
    }
    // In all other cases, look (very hard) for a `xref` parameter.
    reference_in = (name) => {
      var $, base3, ref1;
      if ($ = (ref1 = this.req.variable(name)) != null ? ref1.match(/xref[=:]([\w-]+)/) : void 0) {
        return (base3 = this.session).reference != null ? base3.reference : base3.reference = $[1];
      }
    };
    reference_in('sip_from_params');
    reference_in('sip_to_params');
    reference_in('sip_req_params');
    reference_in('sip_contact_params');
    reference_in('sip_referred_by_params');
    reference_in('sip_h_X-FS-Refer-Params');
    // Otherwise, since the call is coming from a carrier we force the creation of a new context.
    ({Reference} = this.cfg);
    this.reference = new Reference(this.session.reference);
    this.session.reference = this.reference.id;
    // Logger
    // ------
    if ((await this.reference.get_dev_logger())) {
      this.session.dev_logger = true;
    }
    this.direction(C[2]);
    this.notify({
      state: 'incoming-call-carrier-side'
    });
    // SIP Profile
    // -----------

    // Define the (sofia-sip) SIP profiles used to send calls out.
    sip_profile_client = `${pkg.name}-${this.session.profile}-egress`;
    sip_profile_carrier = `${pkg.name}-${this.session.profile}-ingress`;
    this.session.sip_profile = this.req.variable('sip_profile');
    if (this.session.direction === 'ingress') {
      if ((base3 = this.session).sip_profile == null) {
        base3.sip_profile = sip_profile_client;
      }
    } else {
      if ((base4 = this.session).sip_profile == null) {
        base4.sip_profile = sip_profile_carrier;
      }
    }
    this.session.transfer = false;
    // The handled transfer context assumes the transfer request is coming from a (presumably trusted) server. It is used by the tough-rate call-handler.
    // For now this should only happen when a customer calls a global number that points to a conference, and the server that handled the request isn't the one serving the conference.
    this.session.handled_transfer_context = [this.session.profile, 'handled', this.session.reference].join('-');
    if (this.session.direction === 'handled') {
      this.session.transfer = true;
      this.direction('handled');
    }
    // SIP Profile Data
    // ----------------

    // Note that carrier-side the fields are called `sip_profiles` and are stored in the database.
    // The FreeSwitch configuration uses the `profiles` field, which defaults to using port 5080.
    this.session.profile_data = this.cfg.sip_profiles[this.session.profile];
    // Set FreeSwitch variables
    // ------------------------
    sip_params = `xref=${this.session.reference}`;
    our_dialplan = `inline:'socket:127.0.0.1:${(ref1 = this.cfg.port) != null ? ref1 : 5702} async full'`;
    await this.set({
      session_reference: this.session.reference
    });
    await this.export({
      session_reference: this.session.reference,
      // Info for handling of 302 etc. for (I assume) our outbound calls. `cfg.port` is from `thinkable-ducks/server`.
      sip_redirect_profile: this.session.profile,
      sip_redirect_context: this.session.default_transfer_context,
      sip_redirect_dialplan: our_dialplan,
      sip_redirect_contact_params: sip_params,
      sip_invite_params: sip_params,
      sip_invite_to_params: sip_params,
      sip_invite_contact_params: sip_params,
      sip_invite_from_params: sip_params
    });
    debug('Ready', {
      reference: this.session.reference,
      call: this.call.uuid,
      context: this.session.context,
      direction: this.session.direction,
      destination: this.destination,
      source: this.source,
      transfer: this.session.transfer,
      profile: this.session.profile,
      sip_profile: this.session.sip_profile
    });
  };

}).call(this);
