// Generated by CoffeeScript 2.3.2
(function() {
  var SUBSCRIBE, debug, fs, macros, minutes, pkg, seconds, sleep;

  pkg = require('../../package');

  this.name = `${pkg.name}:middleware:client:conference`;

  debug = (require('tangible'))(this.name);

  fs = require('fs');

  ({SUBSCRIBE} = require('red-rings/operations'));

  sleep = function(timeout) {
    return new Promise(function(resolve) {
      return setTimeout(resolve, timeout);
    });
  };

  seconds = 1000;

  minutes = 60 * seconds;

  macros = function(cfg) {
    var __count, count, notify, still_running;
    __count = function(name) {
      return cfg.api(`conference ${name} get count`);
    };
    count = async function(name) {
      var value;
      value = (await __count(name));
      if (value.match(/^\d+/)) {
        return parseInt(value, 10);
      } else {
        return null;
      }
    };
    still_running = async function(name) {
      return ((await count(name))) === !null;
    };
    // Send NOTIFY messages back anytime the conference is updated.
    notify = async function(domain, number) {
      var conf_data, conf_name, content, key, list;
      if (!((domain != null) && (number != null))) {
        return;
      }
      conf_name = `${domain}-conf-${number}`;
      // FIXME only try if local
      list = (await cfg.api(`conference ${conf_name} json_list`).catch(function() {
        return null;
      }));
      if (list == null) {
        return;
      }
      // and list[0] is '['
      content = (function() {
        try {
          return JSON.parse(list);
        } catch (error1) {}
      })();
      if (content == null) {
        return;
      }
      conf_data = content[0];
      if (conf_data == null) {
        return;
      }
      key = `conference:${domain}:${number}`;
      return cfg.rr.notify(key, key, conf_data);
    };
    return {count, still_running, notify};
  };

  this.server_pre = function() {
    var notify;
    ({notify} = macros(this.cfg));
    this.cfg.rr.receive('conference:*').filter(function({op}) {
      return op === SUBSCRIBE;
    }).forEach(function(msg) {
      var $, domain, number, ref;
      if (!($ = (ref = msg.key) != null ? ref.match(/^conference:(\S+):(\d+)$/) : void 0)) {
        return;
      }
      domain = $[1];
      number = $[2];
      return notify(cfg, domain, number);
    });
  };

  this.init = function() {
    this.cfg.record_conference = async(name, key, source) => {
      var last_uri, metadata, recording, still_running, uri;
      ({still_running} = macros(this.cfg));
      debug('conference:record', name);
      recording = (await this.cfg.api(`conference ${name} chkrecord`));
      // Do not start a new recording if one is already active.
      debug('conference:record: recording', name, recording);
      if (!(recording != null ? recording.match(/is not being recorded/) : void 0)) {
        debug('conference:record: Already recording or not ready', name, recording);
        return;
      }
      // Get a URL for recording
      if (this.cfg.recording_uri == null) {
        debug.dev('conference:record: Missing recording_uri', name);
        return;
      }
      metadata = {
        name,
        source,
        conference_start: new Date().toJSON(),
        recording_start: new Date().toJSON()
      };
      uri = (await this.cfg.recording_uri(key, metadata));
      await this.cfg.api(`conference ${name} recording start ${uri}`);
      await this.cfg.api(`conference ${name} play tone_stream://%(125,0,400);%(125,0,450);%(125,0,400)`);
      last_uri = uri;
      while ((await still_running(name))) {
        await sleep(29 * minutes);
        metadata.recording_start = new Date().toJSON();
        uri = (await this.cfg.recording_uri(key, metadata));
        await this.cfg.api(`conference ${name} recording start ${uri}`);
        await this.cfg.api(`conference ${name} play tone_stream://%(125,0,400);%(125,0,450);%(125,0,400)`);
        await sleep(1 * minutes);
        await this.cfg.api(`conference ${name} recording stop ${last_uri}`);
        last_uri = uri;
      }
    };
  };

  this.include = async function() {
    var $, announce, authenticated, conf_name, conf_uri, count, currently, domain, get_conf_pin, id, is_remote, key, language, music_uri, namefile, notify, number, play_in_conference, ref, ref1, res, server, start_recording, uri;
    if (((ref = this.session) != null ? ref.direction : void 0) !== 'conf') {
      return;
    }
    if (this.cfg.host == null) {
      debug.dev('Missing cfg.host');
      return;
    }
    if (this.session.conf == null) {
      debug.dev('Missing conference data');
      return;
    }
    conf_name = this.session.conf.full_name;
    if ($ = conf_name.match(/^(\S+)-conf-(\d+)$/)) {
      domain = $[1];
      number = $[2];
      key = `conference:${domain}:${number}`;
    }
    is_remote = (await this.cfg.is_remote(conf_name, this.session.local_server));
    // Conference is remote
    // --------------------
    if (is_remote) {
      server = is_remote;
      uri = `sip:localconf-${conf_name}@${server};xref={@session.reference}`;
      debug('Conference is remote', uri);
      // We use `deflect` (REFER) because this might happen mid-call (for example inside an IVR menu).
      res = (await this.action('deflect', uri));
      debug('Remote conference returned', uri, res);
      return;
    }
    // Conference is handled locally
    // -----------------------------
    debug('Conference is local');
    this.notify({
      state: 'conference',
      name: conf_name,
      key: key
    });
    conf_uri = (id, name) => {
      return this.prompt.uri('prov', 'prov', id, name);
    };
    id = `number_domain:${this.session.number_domain}`;
    if (this.session.music != null) {
      music_uri = this.session.music;
    }
    if (this.session.conf.music != null) {
      if (this.session.conf.music === false) {
        music_uri = null;
      } else {
        music_uri = conf_uri(id, this.session.conf.music);
      }
    }
    if (music_uri != null) {
      await this.set({
        conference_moh_sound: music_uri
      });
    }
    // Validate passcode if any.
    language = this.session.conf.language;
    if (language == null) {
      language = this.session.language;
    }
    if (language == null) {
      language = this.cfg.announcement_language;
    }
    await this.action('answer');
    await this.set({language});
    await this.sleep(2000);
    get_conf_pin = (o = {}) => {
      return this.prompt.get_pin({
        max: 8,
        tries: 3,
        timeout: 6000,
        file: 'phrase:conference:pin',
        invalid_file: 'phrase:conference:bad_pin'
      });
    };
    authenticated = async() => {
      var customer_pin, pin;
      pin = this.session.conf.pin;
      debug('pin', pin);
      if (pin == null) {
        return true;
      }
      customer_pin = (await get_conf_pin().catch((error) => {
        var ref1;
        debug(`pin error: ${(ref1 = error.stack) != null ? ref1 : error}`);
        return null;
      }));
      return pin === customer_pin;
    };
    if ((await authenticated())) {
      namefile = `/tmp/${this.session.logger_uuid}-name.wav`;
      // This uses `playback`, but `@action 'phrase', 'voicemail_record_name'` (separator is `,` for parameters) should work as well.
      await this.action('playback', 'phrase:voicemail_record_name');
      debug('record');
      await this.action('record', `${namefile} 2`);
      // Announce number of persons in conference
      // ----------------------------------------
      ({count, notify} = macros(this.cfg));
      currently = (await count(conf_name));
      if (currently == null) {
        currently = 0;
      }
      await this.action('playback', `phrase:conference:count:${currently}`);
      // Play in conference
      // ------------------

      // The thing, really, is that conference uses `switch_core_file` and parses for `say:` only, while `playback` (in `mod_dptools`) uses `switch_ivr_play_file`, which parses `phrase:`, `say:` etc.
      // Really we should just barge on the channel if we need anything more complex than playing files, tone-streams, etc.
      play_in_conference = (what) => {
        return this.call.api(['conference', conf_name, 'play', what].join(' ')); // [conference API commands](https://freeswitch.org/confluence/display/FREESWITCH/mod_conference#mod_conference-APIReference)
      };
      announce = async() => {
        debug('announce');
        await play_in_conference('tone_stream://%(125,0,300);%(125,0,450);%(125,0,600)');
        return (await play_in_conference(namefile).catch((error) => {
          var ref1;
          return debug(`error: ${(ref1 = error.stack) != null ? ref1 : error}`);
        }));
      };
      this.call.once('cleanup_linger', async() => {
        await play_in_conference('tone_stream://%(125,0,600);%(125,0,450);%(125,0,300)');
        await play_in_conference(namefile);
        try {
          // FIXME: This assumes we are co-located (inside the same container) as FreeSwitch
          // which is stronger than our regular assumption (=we are co-located on the same VM
          // but not in the same container).
          // We should ask FreeSwitch to do the `unlink` here.
          fs.unlinkSync(namefile);
        } catch (error1) {}
        return notify(domain, number);
      });
      setTimeout(announce, 1000);
      await this.set({
        conference_max_members: (ref1 = this.session.conf.max_members) != null ? ref1 : null
      });
      // Log into the conference
      debug('conference');
      await this.reference.set_number_domain(this.session.number_domain);
      this.notify({
        state: 'conference:started',
        name: conf_name,
        key: key
      });
      notify(domain, number);
      // * doc.number_domain.conferences[].record (boolean) If true the conference calls will be recorded.
      if (this.session.conf.record) {
        start_recording = () => {
          return this.cfg.record_conference(conf_name, key, this.source);
        };
        setTimeout(start_recording, 1000);
      }
      await this.action('conference', `${conf_name}++flags{}`);
    }
  };

}).call(this);
