// Generated by CoffeeScript 2.3.2
(function() {
  // Internally we route french numbers like any international number,
  // e.g. 33+national number (without the "0" or "E" prefix)
  // however in France the dialing plan cannot be readily mapped into
  // the international numbering plan because special codes, etc. in the
  // national dialing plan do interfere with national prefixes.
  // (For example, 112 and 3615 are prefixes for geographic numbers.)

  // On the other hand, in order to be routable all output numbers must
  // contain digits. We use `_` (underscore) to prefix special numbers.

  // For example:
  // * 3615 (french national dialing plan) is mapped to 33_3615 (international numbering plan)
  // * 112 (french national dialing plan) is mapped to 33_112 (international numbering plan)

  // See http://www.arcep.fr/index.php?id=interactivenumeros
  var debug, pkg, translate_source;

  pkg = require('../../../package.json');

  this.name = `${pkg.name}:middleware:client:egress:national-FR`;

  debug = (require('tangible'))(this.name);

  translate_source = function(source) {
    var $;
    switch (false) {
      // From: national number
      case !($ = source.match(/^(0|\+33)([123456789][0-9]{8})$/)):
        return `33${$[2]}`;
      // From: international number (why??)
      case !($ = source.match(/^(00|\+)([2-9][0-9]*)$/)):
        return $[2];
      default:
        debug(`Cannot translate source ${source}`);
        return null;
    }
  };

  this.include = function() {
    var entry, i, len, m, new_asserted, new_source, patterns, ref;
    if (((ref = this.session) != null ? ref.direction : void 0) !== 'egress') {
      return;
    }
    if (this.session.dialplan !== 'national') {
      return;
    }
    if (this.session.country !== 'fr') {
      return;
    }
    debug('source', this.source);
    // Verify that the caller-id follows the proper format
    // ---------------------------------------------------
    new_source = translate_source(this.source);
    if (new_source != null) {
      this.session.ccnq_from_e164 = new_source;
    }
    if (this.session.asserted != null) {
      new_asserted = translate_source(this.session.asserted);
      if (new_asserted != null) {
        this.session.asserted = new_asserted;
      }
    }
    // Verify that the called number follows the proper format
    // -------------------------------------------------------

    // International numbers embedded inside the National numbering plan:
    // decision ARCEP 06-0720
    patterns = [
      {
        // ARCEP 06-0720

        // fixes mayotte, réunion, territoires océan indien
        match: /^(0|\+33)(26[29].*)$/,
        now: function($) {
          return `262${$[2]}`;
        }
      },
      {
        // mobiles mayotte
        match: /^(0|\+33)(639.*)$/,
        now: function($) {
          return `262${$[2]}`;
        }
      },
      {
        // decision ARCEP 06-0535 + 00-0536

        // mobiles-guadeloupe
        match: /^(0|\+33)(69[01].*)$/,
        now: function($) {
          return `590${$[2]}`;
        }
      },
      {
        // mobiles-reunion
        match: /^(0|\+33)(69[23].*)$/,
        now: function($) {
          return `262${$[2]}`;
        }
      },
      {
        // mobiles-guyane
        match: /^(0|\+33)(694.*)$/,
        now: function($) {
          return `594${$[2]}`;
        }
      },
      {
        // mobiles-martinique
        match: /^(0|\+33)(69[67].*)$/,
        now: function($) {
          return `596${$[2]}`;
        }
      },
      {
        // other, fixes
        // fixes-guadeloupe
        match: /^(0|\+33)(590.*)$/,
        now: function($) {
          return `590${$[2]}`;
        }
      },
      {
        // fixes-reunion
        match: /^(0|\+33)(262.*)$/,
        now: function($) {
          return `262${$[2]}`;
        }
      },
      {
        // fixes-guyane
        match: /^(0|\+33)(594.*)$/,
        now: function($) {
          return `594${$[2]}`;
        }
      },
      {
        // fixes-martinique
        match: /^(0|\+33)(596.*)$/,
        now: function($) {
          return `596${$[2]}`;
        }
      },
      {
        // actually fixes and mobiles are mixed. Also the international dialplan is different, see http://www.itu.int/oth/T02020000B2/en
        // fixes-stpierre
        match: /^(0|\+33)(508.*)$/,
        now: function($) {
          return $[2];
        }
      },
      {
        // ARCEP 04-0847
        // nongeo-guadeloupe
        match: /^(0|\+33)(876[01].*|976[018])$/,
        now: function($) {
          return `590${$[2]}`;
        }
      },
      {
        // nongeo-reunion
        match: /^(0|\+33)(876[23].*|976[239])$/,
        now: function($) {
          return `262${$2}`;
        }
      },
      {
        // nongeo-guyane
        match: /^(0|\+33)(876[4].*|976[45])$/,
        now: function($) {
          return `594${$[2]}`;
        }
      },
      {
        // nongeo-martinique
        match: /^(0|\+33)(876[67].*|976[67])$/,
        now: function($) {
          return `596${$[2]}`;
        }
      },
      {
        // geographic, non-geographic, mobile, or service
        match: /^(0|\+33)([123456789][0-9]{8})$/,
        now: function($) {
          return `33${$[2]}`;
        }
      },
      {
        // special services
        match: /^(1[0-9]{1,5}|3[0-9]{3})$/,
        now: function($) {
          return `33_${$[1]}`;
        }
      },
      {
        // international call
        match: /^(00|\+)([0-9]*)$/,
        now: function($) {
          return $[2];
        }
      }
    ];
    for (i = 0, len = patterns.length; i < len; i++) {
      entry = patterns[i];
      m = this.destination.match(entry.match);
      if (m != null) {
        this.session.ccnq_to_e164 = entry.now(m);
        debug('Found', {
          to_e164: this.session.ccnq_to_e164
        });
        return;
      }
    }
    return debug('None found', {
      destination: this.destination
    });
  };

}).call(this);
