// Generated by CoffeeScript 2.3.2
(function() {
  var debug, default_music, default_ringback, pkg, set_params, tones, url;

  pkg = require('../../../package.json');

  this.name = `${pkg.name}:middleware:client:ingress:post`;

  debug = (require('tangible'))(this.name);

  url = require('url');

  tones = require('../tones');

  // Use fr-ring for default ringback
  default_ringback = tones.fr.ringback;

  // See https://freeswitch.org/jira/browse/FS-9776
  default_music = 'silence';

  // Call-Handler
  // ============
  this.include = async function() {
    var base, base1, base2, base3, base4, caller, converged, convergence, dst_number, i, len, list, list_id, name, parameters, pid, ref, ref1, to_uri;
    if (((ref = this.session) != null ? ref.direction : void 0) !== 'ingress') {
      return;
    }
    debug('Ready', {
      dialplan: this.session.dialplan,
      country: this.session.country,
      destination: this.destination,
      number_domain: this.session.number_domain
    });
    if (this.session.number_domain == null) {
      debug.dev('Missing number_domain');
      return;
    }
    // Routing
    // -------

    // One of the national translations should have mapped us to a different dialplan (e.g. 'national').
    if (this.session.dialplan === 'e164') {
      return this.respond('484');
    }
    // Retrieve number data.
    dst_number = (await this.validate_local_number());
    if (dst_number == null) {
      debug('Number not found');
      return;
    }
    // Call rejection: reject anonymous caller

    // * doc.local_number.reject_anonymous (boolean) If true, rejects anonymous calls.
    // * session.caller_privacy (boolean) Indicates if the caller requested privacy.
    // * doc.local_number.reject_anonymous_to_voicemail (boolean) If true, rejected anonymous calls are sent to voicemail. If false, they are rejected with an error message.
    // * doc.config%3Avoice_prompts/reject-anonymous.wav The attachment played when an inbound anonymous call is rejected and the call is not sent to voicemail.
    if (this.session.number.reject_anonymous) {
      if (this.session.caller_privacy) {
        if (this.session.number.reject_anonymous_to_voicemail) {
          debug('reject anonymous: send to voicemail');
          // @destination unchanged
          this.direction('voicemail');
          return;
        }
        debug('reject anonymous');
        this.notify({
          state: 'reject-anonymous'
        });
        // return @respond '603 Decline (anonymous)'
        await this.action('answer');
        // `provisioning` is a `nimble-direction` convention.
        await this.action('playback', `${this.cfg.provisioning}/config%3Avoice_prompts/reject-anonymous.wav`);
        return this.action('hangup');
      }
    }
    // * doc.local_number.use_blacklist (boolean) If true and a `list:<destination-number>@<calling-number>` record exists, where `<destination-number>` is the identifier of a local-number (in the format `<number>@<number-domain>`), use that record to decide whether to reject the inbound call based on the calling number.
    // * doc.local_number.use_whitelist (boolean) If true and a `list:<destination-number>@<calling-number>` record exists, where `<destination-number>` is the identifier of a local-number (in the format `<number>@<number-domain>`), use that record to decide whetehr to accept the inbound call.
    // * doc.list A record indicating whether to accept a inbound call or reject it. The identifier is `list:<number>@<number-domain>@<caller>` with `<number>` and `<caller>` interpreted in the given number-domain.
    // * doc.list.disabled (boolean) If true, proceed as-if the record did not exist.
    // * doc.list.blacklist (boolean) If true, the call is rejected.
    // * doc.list.whitelist (boolean) If false, the call is rejected.
    // * doc.local_number.list_to_voicemail (boolean) If true, calls rejected by blacklist or whitelist are forwarded to voicemail. Otherwise calls are rejected without any error message. Default: false.
    if (this.session.number.use_blacklist || this.session.number.use_whitelist) {
      pid = this.req.header('P-Asserted-Identity');
      caller = pid != null ? url.parse(pid).auth : this.source;
      list_id = `list:${dst_number}@${caller}`;
      debug(`Number ${dst_number}, requesting caller ${caller} list ${list_id}`);
      list = (await this.cfg.prov.get(list_id).catch(function() {
        return {};
      }));
      if (!list.disabled) {
        if (this.session.number.use_blacklist && list.blacklist) {
          this.notify({
            state: 'blacklisted'
          });
          if (this.session.number.list_to_voicemail) {
            // @destination unchanged
            this.direction('voicemail');
            return;
          }
          return this.respond('486 Decline (blacklisted)'); // was 603
        }
        if (this.session.number.use_whitelist && !list.whitelist) {
          this.notify({
            state: 'not whitelisted'
          });
          if (this.session.number.list_to_voicemail) {
            // @destination unchanged
            this.direction('voicemail');
            return;
          }
          return this.respond('486 Decline (not whitelisted)'); // was 603
        }
      }
    }
    
    // * doc.local_number.custom_ringback (boolean,string) If present, a custom ringback is played while the call is being presented to the destination user. The ringback file is an attachment located in `doc.voicemail_settings`; its name is the value of `custom_ringback`, or `ringback.wav` if `custom_ringback` is `true`. Default: plays system-wide `cfg.ringback`, or a code-assigned default ringback.
    debug('ringback', this.session.ringback, this.session.number.custom_ringback);
    if (this.session.number.custom_ringback === true) {
      if ((base = this.session).ringback == null) {
        base.ringback = [this.cfg.userdb_base_uri, this.session.number.user_database, 'voicemail_settings', 'ringback.wav'].join('/');
      }
    }
    if (typeof this.session.number.custom_ringback === 'string') {
      if ((base1 = this.session).ringback == null) {
        base1.ringback = [this.cfg.userdb_base_uri, this.session.number.user_database, 'voicemail_settings', this.session.number.custom_ringback].join('/');
      }
    }
    // * doc.local_number.custom_music (boolean,string) If present, a custom music is played while the call is put on-hold. The music-on-hold file is an attachment located in `doc.voicemail_settings`; its name is the value of `custom_music`, or `music.wav` if `custom_ringback` is `true`. Default: plays system-wide `cfg.music`, or a code-assigned default music (bips).
    if (this.session.number.custom_music === true) {
      if ((base2 = this.session).music == null) {
        base2.music = [this.cfg.userdb_base_uri, this.session.number.user_database, 'voicemail_settings', 'music.wav'].join('/');
      }
    }
    if (typeof this.session.number.custom_music === 'string') {
      if ((base3 = this.session).music == null) {
        base3.music = [this.cfg.userdb_base_uri, this.session.number.user_database, 'voicemail_settings', this.session.number.custom_music].join('/');
      }
    }
    // * cfg.answer_for_ringback (boolean) If true, answer the call (200 OK) instead of pre-answering the call (183 with Media) for custom ringback.
    // * session.answer_for_ringback (boolean) If true, answer the call (200 OK) instead of pre-answering the call (183 with Media) for custom ringback.
    // * cfg.ready_for_ringback (boolean) If true, inbound calls are ring-ready (180 without media) immediately, without waiting for the customer device to provide ringback.
    // * session.ready_for_ringback (boolean) If true, inbound calls are ring-ready (180 without media) immediately, without waiting for the customer device to provide ringback.
    // * doc.local_number.ring_ready (boolean) If true, inbound calls are ring-ready (180 without media) immediately, without waiting for the customer device to provide ringback.
    debug('Ringback');
    if (this.session.number.custom_ringback) {
      if (this.cfg.answer_for_ringback || this.session.answer_for_ringback) {
        debug('answer for ringback');
        await this.action('answer'); // 200
        this.session.sip_wait_for_aleg_ack = false;
      } else {
        debug('pre_answer for ringback');
        await this.action('pre_answer'); // 183
      }
    } else {
      if (this.session.cf_active || this.cfg.ready_for_ringback || this.session.ready_for_ringback || this.session.number.ring_ready) {
        debug('cf_active');
        await this.action('ring_ready'); // 180
      }
    }
    
    // So far we have no reason to reject the call.
    await set_params.call(this);
    // `CF...` can be either configured as URIs (number.cfa etc. -- bypasses controls) or as plain numbers (will use the `forward` direction for access control).

    // * session.cf_active (boolean) true if any type of call forwarding is present on the local-number
    // * doc.local_number.cfa_enabled (boolean) If false, Call Forward All is disabled.
    // * doc.local_number.cfb_enabled (boolean) If false, Call Forward on Busy is disabled.
    // * doc.local_number.cfnr_enabled (boolean) If false, Call Forward on Not Registered is disabled.
    // * doc.local_number.cfda_enabled (boolean) If false, Call Forward on Don't Answer is disabled.
    // * doc.local_number.cfa_voicemail (boolean) If true, all incoming calls are sent to voicemail.
    // * doc.local_number.cfa_number (string) If present and the call is not sent to voicemail, all incoming calls are sent to this number (interpreted in the local number-domain).
    // * doc.local_number.cfa (string:URI) If present and the call is not sent to voicemail or forward to a number, all incoming calls are sent to this URI.
    // * doc.local_number.cfb_voicemail (boolean) If true, incoming Busy calls are sent to voicemail.
    // * doc.local_number.cfb_number (string) If present and the call is not sent to voicemail, incoming Busy calls are sent to this number (interpreted in the local number-domain).
    // * doc.local_number.cfb (string:URI) If present and the call is not sent to voicemail or forward to a number, incoming Busy calls are sent to this URI.
    // * doc.local_number.cfnr_voicemail (boolean) If true, incoming Not Registered calls are sent to voicemail.
    // * doc.local_number.cfnr_number (string) If present and the call is not sent to voicemail, incoming Not Registered calls are sent to this number (interpreted in the local number-domain).
    // * doc.local_number.cfnr (string:URI) If present and the call is not sent to voicemail or forward to a number, incoming Not Registered calls are sent to this URI.
    // * doc.local_number.cfda_voicemail (boolean) If true, incoming Don't Answer calls are sent to voicemail.
    // * doc.local_number.cfda_number (string) If present and the call is not sent to voicemail, incoming Don't Answer calls are sent to this number (interpreted in the local number-domain).
    // * doc.local_number.cfda (string:URI) If present and the call is not sent to voicemail or forward to a number, incoming Don't Answer calls are sent to this URI.
    this.session.cf_active = false;
    ref1 = ['cfa', 'cfb', 'cfnr', 'cfda'];
    for (i = 0, len = ref1.length; i < len; i++) {
      name = ref1[i];
      ((name) => {
        var n, p, v;
        if (this.session.number[`${name}_enabled`] === false) {
          return;
        }
        v = this.session[`${name}_voicemail`] = this.session.number[`${name}_voicemail`];
        n = this.session[`${name}_number`] = this.session.number[`${name}_number`];
        p = this.session[name] = this.session.number[name];
        this.session.cf_active = this.session.cf_active || (v != null) || (n != null) || (p != null);
        // Make sure we get the messages even if the call is forwarded.
        return this.session[`${name}_voicemail_number`] = this.destination;
      })(name);
    }
    // Call Forward All
    // ----------------

    // * session.reason (string) The RFC5806 `reason` field for call forwarding.
    debug('CFA?');
    this.session.reason = 'unconditional'; // RFC5806
    if (this.session.cfa_voicemail) {
      debug('cfa:voicemail');
      this.notify({
        state: 'cfa:voicemail'
      });
      this.destination = this.session.cfa_voicemail_number;
      this.direction('voicemail');
      return;
    }
    if (this.session.cfa_number != null) {
      debug('cfa:forward');
      this.notify({
        state: 'cfa:forward'
      });
      this.session.destination = this.session.cfa_number;
      this.direction('forward');
      return;
    }
    if (this.session.cfa != null) {
      debug('cfa:fallback');
      this.notify({
        state: 'cfa:fallback'
      });
      this.session.initial_destinations = [
        {
          to_uri: this.session.cfa
        }
      ];
      return;
    }
    this.session.reason = null;
    // Do Not Disturb
    // --------------

    // * local_number.dnd (boolean) If true, considers the line is in Do Not Disturb and use the CFB indications to handle the call.
    debug('DND?');
    if (this.session.number.dnd) {
      this.session.reason = 'do-not-disturb'; // RFC5806
      if (this.session.cfb_voicemail) {
        debug('dnd:voicemail');
        this.notify({
          state: 'dnd:voicemail'
        });
        this.destination = this.session.cfb_voicemail_number;
        this.direction('voicemail');
        return;
      }
      if (this.session.cfb_number != null) {
        debug('dnd:forward');
        this.notify({
          state: 'dnd:forward'
        });
        this.session.destination = this.session.cfb_number;
        this.direction('forward');
        return;
      }
      if (this.session.cfb != null) {
        debug('dnd:fallback');
        this.notify({
          state: 'dnd:fallback'
        });
        this.session.initial_destinations = [
          {
            to_uri: this.session.cfb
          }
        ];
        return;
      }
      this.session.reason = null;
    }
    // Build the destination FreeSwitch dialstring
    // -------------------------------------------

    // Note the different alternatives for routing:
    // - To URI: `sofia/.../<To-URI>` (and RURI if the RURI is not specified)
    // - RURI: `sip_invite_req_uri`
    // - Route header: `sip_route_uri` = `sip:<domain-name>`
    // - Network destination: `sip_network_ip`, 'sip_network_port'
    // - `;fs_path=`

    // ### Standard destination
    parameters = [];
    to_uri = `sip:${this.session.endpoint_name // set by validate_local_number()
}`;
    
    // * cfg.ingress_target (string:domain) Inbound domain for static endpoint: inbound calls are sent to the specified domain if the endpoint's name does not contain `@`.
    if (!to_uri.match(/@/)) {
      to_uri = `sip:${this.destination}@${this.cfg.ingress_target}`;
    }
    // * doc.local_number.endpoint_via (string:domain) If present, inbound calls are sent to the specified server instead of the endpoint's.
    if (this.session.number.endpoint_via != null) {
      //# A proper way to do it:
      // parameters.push "sip_route_uri=sip:#{@session.number.endpoint_via}"
      //# How we've been doing it:
      to_uri = `sip:${this.destination}@${this.session.number.endpoint_via}`;
    }
    // Convergence
    // -----------

    // * doc.local_number.convergence_active (boolean) If true, the convergence feature is active.
    // * doc.local_number.convergence (array of objects) Convergence destinations and options.
    // * doc.local_number.convergence[].number (string) destination number (as dialed from the endpoint)
    // * doc.local_number.convergence[].confirm (boolean) whether to ask for confirmation after the call is answered
    // * doc.local_number.convergence[].confirm (string) whether to ask for confirmation after the call is answered, and which digit should be used to confirm
    // * doc.local_number.convergence[].delay (integer) number of seconds to wait before dialing this destination
    // * doc.local_number.convergence[].timeout (integer) number of seconds to wait before considering the destination as not responding

    // The convergence function returns a list of optional, additional targets (e.g. mobile phone destinations) which are called at the same time as the original number, or with a slight delay. This is used to implement "Follow-Me"/"Mobile Convergence"-type scenarios.
    convergence = async() => {
      var confirm, default_params, host, key, language, p, port, profile, ref2, ref3, ref4;
      // If the feature is not enabled on this line just skip.
      if (!((this.session.number.convergence != null) && this.session.number.convergence_active)) {
        return [];
      }
      // Following code lifted from place-call (esp. the conference code).

      // The additional calls will be sent back to ourselves, we need to figure out our host and port.
      if (!((((ref2 = this.cfg.session) != null ? ref2.profile : void 0) != null) && (this.cfg.profiles != null))) {
        return [];
      }
      profile = this.cfg.session.profile;
      ({host} = this.cfg);
      p = this.cfg.profiles[profile];
      if (!(host && (p != null))) {
        return [];
      }
      port = (ref3 = p.egress_sip_port) != null ? ref3 : p.sip_port + 10000;
      // Call confirmation
      confirm = this.session.number.convergence_confirm;
      if (confirm != null) {
        // The `confirm` field can be a boolean or a string.
        key = '5';
        if (typeof confirm === 'string' && confirm.match(/^\d$/)) {
          key = confirm;
        }
        // Try hard to figure out what language we should use.
        language = this.session.language;
        if (language == null) {
          language = this.session.number.language;
        }
        if (language == null) {
          language = this.cfg.announcement_language;
        }
        if (language == null) {
          language = '';
        }
        // The `call_options` are used by tough-rate.
        await this.reference.set_call_options({
          group_confirm_key: key,
          group_confirm_file: `phrase:confirm:${key}`,
          group_confirm_error_file: `phrase:confirm:${key}`,
          group_confirm_read_timeout: 15000, // defaults to 5000
          group_confirm_cancel_timeout: false,
          language: language
        });
      }
      default_params = {};
      // Normally the x-ref parameters are already defined in `middleware/client/setup`.
      // (The `call-to-conference` function needs to define them but we should not need to.)

      // xref = "xref:#{@session.reference}"
      // default_param.sip_invite_params = xref
      // default_param.sip_invite_to_params = xref
      // default_param.sip_invite_contact_params = xref
      // default_param.sip_invite_from_params = xref

      // This is similar to what is done in `middleware/forward/basic`: we override the calling number if requested to.
      if (this.cfg.mask_source_on_forward) {
        default_params.origination_caller_id_number = (ref4 = this.session.number.asserted_number) != null ? ref4 : this.session.number.number;
      }
      // Define parameters and targets.
      return this.session.number.convergence.filter(function(o) {
        return o.number != null;
      }).map((o) => {
        var params;
        params = Object.assign({}, default_params);
        // Delay ("Follow-Me" application)
        if (o.delay) {
          params.leg_delay_start = o.delay;
        }
        // Timeout
        if (o.timeout) {
          params.leg_timeout = o.timeout;
        }
        return {
          parameters: Object.keys(params).map(function(k) {
            return `${k}=${params[k]}`;
          }),
          to_uri: `sip:${o.number}@${host}:${port}`
        };
      });
    };
    converged = (await convergence());
    if ((base4 = this.session).initial_destinations == null) {
      base4.initial_destinations = [{parameters, to_uri}, ...converged];
    }
    this.notify({
      state: 'ingress-call',
      endpoint: this.session.endpoint_name
    });
    await this.reference.set_endpoint(this.session.endpoint_name);
    // ### Build the set of `_in` targets for notifications of the reference data.
    if (this.session.dev_logger) {
      await this.reference.set_dev_logger(true);
    }
    if (this.session.number.record_ingress) {
      this.record_call(this.session.number._id);
    }
    debug('Done.');
  };

  // `set_params`
  // ============

  // Non-call-handling-specific parameters (these are set on all calls independently of call treatment).
  set_params = async function() {
    var base, base1, base2, base3, base4, base5, base6, dlg_timeout, fr_inv_timeout, fr_timeout, ref, ref1, ref2, ref3;
    debug('set_params');
    if ((this.session.country != null) && this.session.country in tones) {
      if ((base = this.session).ringback == null) {
        base.ringback = tones[this.session.country].ringback;
      }
      if (this.cfg.use_country_tones_for_music) {
        if ((base1 = this.session).music == null) {
          base1.music = tones.loop(tones[this.session.country].waiting);
        }
      }
    }
    if ((base2 = this.session).ringback == null) {
      base2.ringback = this.cfg.ringback;
    }
    if ((base3 = this.session).ringback == null) {
      base3.ringback = default_ringback;
    }
    if ((base4 = this.session).music == null) {
      base4.music = this.cfg.music;
    }
    if ((base5 = this.session).music == null) {
      base5.music = default_music;
    }
    if ((base6 = this.session).sip_wait_for_aleg_ack == null) {
      base6.sip_wait_for_aleg_ack = true;
    }
    debug('set_params', {
      ringback: this.session.ringback,
      music: this.session.music,
      sip_wait_for_aleg_ack: this.session.sip_wait_for_aleg_ack
    });
    await this.set({
      ccnq_endpoint: this.session.endpoint_name
    });
    // * doc.local_number.dialog_timer (number) Maximum duration of a call for this local-number.
    // * doc.local_number.inv_timer (number) Maximum progress duration for this local-number. Typically this is the duration before the call is sent to voicemail.
    dlg_timeout = (ref = this.session.number.dialog_timer) != null ? ref : 28000; // 8h
    fr_inv_timeout = (ref1 = this.session.number.inv_timer) != null ? ref1 : 90;
    fr_timeout = (ref2 = this.session.number.timer) != null ? ref2 : 2; // Unused
    
    // Maximal call duration

    // Note: tough-rate uses `dialog_timeout` for this (which isn't on the wiki).
    debug('schedule hangup');
    await this.action('sched_hangup', `+${dlg_timeout}`);
    this.session.cdr_direction = this.session.direction;
    debug('set parameters');
    await this.set({
      // These are injected so that they may eventually show up in CDRs.

      // * doc.local_number.account (string) Account information for this number. Normally not present, since the account information from the endpoint is used. Overrides the endpoint's account information if present.
      ccnq_direction: this.session.direction,
      ccnq_account: this.session.number.account,
      ccnq_profile: this.session.profile,
      ccnq_from_e164: this.session.ccnq_from_e164,
      ccnq_to_e164: this.session.ccnq_to_e164,
      // Other SIP parameters

      // [progress timeout = PDD](https://wiki.freeswitch.org/wiki/Channel_Variables#progress_timeout)
      // counts from the time the INVITE is placed until a progress indication (e.g. 180, 183) is received. Controls Post-Dial-Delay on this leg.
      progress_timeout: 18,
      // [call timeout = useless, use originate timeout or leg timeout](https://wiki.freeswitch.org/wiki/Channel_Variables#call_timeout)
      call_timeout: 300,
      sip_contact_user: this.session.ccnq_from_e164,
      effective_caller_id_number: this.source,
      sip_cid_type: 'pid',
      'sip_h_X-En': this.session.endpoint_name,
      // These should not be forwarded towards customers.
      'sip_h_X-At': null,
      'sip_h_X-Ex': null,
      'sip_h_X-RH': null,
      'sip_h_X-RP': null,
      'sip_h_X-RR': null,
      'sip_h_X-RT': null,
      'sip_h_X-RU': null,
      // Ringbacks
      ringback: this.session.ringback, // Used for pre_answer
      instant_ringback: false,
      transfer_ringback: this.session.ringback, // Used after answer
      
      // Codec negotiation with late-neg:
      inherit_codec: (ref3 = this.session.inherit_codec) != null ? ref3 : true
    });
    // * hdr.X-En Set on inbound calls to the endpoint of the local-number.
    debug('export parameters');
    await this.export({
      t38_passthru: true,
      sip_wait_for_aleg_ack: this.session.sip_wait_for_aleg_ack,
      'sip_h_X-En': this.session.endpoint_name,
      originate_timeout: fr_inv_timeout,
      bridge_answer_timeout: fr_inv_timeout,
      // Music
      hold_music: this.session.music
    });
    debug('set_params: done.');
  };

}).call(this);
