// Generated by CoffeeScript 2.3.2
(function() {
  var debug;

  this.name = 'huge-play:middleware:dtmf';

  debug = (require('tangible'))(this.name);

  this.include = async function() {
    var dtmf_buffer, expect, handler, on_change, present, timer, wait_for_digit;
    timer = null;
    handler = null;
    dtmf_buffer = '';
    on_change = null;
    await this.call.event_json('DTMF');
    this.call.on('DTMF', (res) => {
      var c;
      c = res.body['DTMF-Digit'];
      debug(`Received ${c}`);
      if (dtmf_buffer == null) {
        dtmf_buffer = '';
      }
      dtmf_buffer += c;
      if (typeof on_change === "function") {
        on_change();
      }
    });
    wait_for_digit = (timeout) => {
      if (on_change != null) {
        debug.dev("wait_for_digit called while already active (ignore)");
        return Promise.reject(new Error('recursive'));
      }
      return new Promise((resolve, reject) => {
        var on_timeout;
        on_timeout = () => {
          on_change = null;
          return reject(new Error('timeout'));
        };
        timer = setTimeout(on_timeout, timeout);
        on_change = function() {
          on_change = null;
          clearTimeout(timer);
          return resolve();
        };
      });
    };
    expect = async(min_length = 1, max_length = min_length, inter_digit = 3 * 1000) => {
      var clear, error, m;
      debug('expect', min_length, max_length, inter_digit);
      clear = function(r = dtmf_buffer) {
        debug('clear');
        dtmf_buffer = '';
        if (r.length < min_length) {
          return '';
        } else {
          return r.substr(0, max_length);
        }
      };
      while (true) {
        // If the user already provide a value, terminated with `#`,
        if (m = dtmf_buffer.match(/^([^#]*)#/)) {
          // then use that value.
          return clear(m[1]);
        }
        // Make sure we don't overflow.
        if (dtmf_buffer.length >= max_length) {
          return clear();
        }
        try {
          // Wait for a digit
          await wait_for_digit(inter_digit);
        } catch (error1) {
          // If we didn't collect a digit,
          error = error1;
          // do we already have enough?
          return clear();
        }
      }
    };
    present = function() {
      return dtmf_buffer.length > 0;
    };
    // Public API
    // ----------
    this.dtmf = {
      // `@dtmf.clear`: resets all fields. If the buffer was not empty, returns the content of the buffer.
      clear: function() {
        var r;
        [r, dtmf_buffer] = [dtmf_buffer, ''];
        return r;
      },
      // `@dtmf.expect min, max, inter_digit`: Returns a Promise that will resolve once at least `min` (default: 1) and at most `max` (default: `min`) digits have been received, with an inter-digit timeout of `inter_digit` (in ms, default: 3000). If the number of collected digits is not enough, return the empty string.
      expect: expect,
      // `@dtmf.playback`: execute a playback command in FreeSwitch, unless a digit has already been entered.
      playback: async(url) => {
        if (present()) {
          return;
        }
        await this.set({
          playback_terminators: '1234567890#*'
        });
        return (await this.action('playback', url));
      },
      // `@dtmf.phrase`: execute a phrase command in FreeSwitch, unless a digit has already been entered.
      phrase: async(phrase) => {
        if (present()) {
          return;
        }
        await this.set({
          playback_terminators: '1234567890#*'
        });
        return (await this.action('phrase', phrase));
      }
    };
    // Typical pattern is:
    // ```
    // # start menu, clear DTMF buffer
    // @dtmf.clear()
    // # play the prompt
    // await @dtmf.playback prompt_file
    // # expect between one and two digits
    // switch await @dtmf.expect 1, 2
    //   when '1'
    //   when '2'
    // ```
    return null;
  };

}).call(this);
