// Generated by CoffeeScript 2.3.2
(function() {
  var BlueRingReference, GET, RedisInterfaceReference, Reference, SET, Solid, reference_id, seconds;

  Solid = require('solid-gun');

  seconds = 1000;

  reference_id = function() {
    return Solid.time() + Solid.uniqueness();
  };

  SET = function(name) {
    return function(value) {
      return this.set(name, value);
    };
  };

  GET = function(name) {
    return function() {
      return this.get(name);
    };
  };

  Reference = (function() {
    class Reference {
      // This needs to be extended with ::timeout and ::interface (a blue-rings or redis instance).
      _key(name) {
        return `xref-${this.id}.${name}`;
      }

      constructor(id) {
        if (id == null) {
          id = reference_id();
        }
        this.id = id;
      }

    };

    Reference.prototype.set_endpoint = SET('endpoint');

    Reference.prototype.set_destination = SET('destination');

    Reference.prototype.set_source = SET('source');

    Reference.prototype.set_domain = SET('domain');

    Reference.prototype.set_dev_logger = SET('dev_logger');

    Reference.prototype.set_account = SET('account');

    Reference.prototype.set_call_options = SET('call_options');

    Reference.prototype.set_block_dtmf = SET('block_dtmf');

    Reference.prototype.set_number = SET('number');

    Reference.prototype.set_number_domain = SET('number_domain');

    Reference.prototype.get_endpoint = GET('endpoint');

    Reference.prototype.get_destination = GET('destination');

    Reference.prototype.get_source = GET('source');

    Reference.prototype.get_domain = GET('domain');

    Reference.prototype.get_dev_logger = GET('dev_logger');

    Reference.prototype.get_account = GET('account');

    Reference.prototype.get_call_options = GET('call_options');

    Reference.prototype.get_block_dtmf = GET('block_dtmf');

    Reference.prototype.get_number = GET('number');

    Reference.prototype.get_number_domain = GET('number_domain');

    return Reference;

  }).call(this);

  // Reference, using a RedisInterface (from `normaly-key/interface`) as its `interface`.
  RedisInterfaceReference = class RedisInterfaceReference extends Reference {
    set(name, value) {
      var key;
      key = this._key(name);
      return this.interface.set(key, name, value);
    }

    get(name) {
      var key;
      key = this._key(name);
      return this.interface.get(key, name);
    }

    clear_(cat) {
      var key;
      key = this._key(cat);
      return this.interface.clear(key);
    }

    add_(cat, value) {
      var key;
      key = this._key(cat);
      return this.interface.add(key, value);
    }

    del_(cat, value) {
      var key;
      key = this._key(cat);
      return this.interface.remove(key, value);
    }

    has_(cat, value) {
      var key;
      key = this._key(cat);
      return this.interface.has(key, value);
    }

    all_(cat) {
      var key;
      key = this._key(cat);
      return this.interface.members(key);
    }

  };

  // Reference, using a BlueRing instance (from `blue-rings`) as its `interface`.
  BlueRingReference = class BlueRingReference extends Reference {
    _expiry() {
      return Date.now() + this.timeout * seconds;
    }

    set(name, value) {
      var key;
      key = this._key(name);
      this.interface.setup_text(key, this._expiry());
      this.interface.update_text(key, value);
    }

    get(name) {
      var coherent, key, value;
      key = this._key(name);
      [coherent, value] = this.interface.get_text(key);
      return value;
    }

    // Clearable sets
    // --------------

    // Currently we use three such sets: 'skill', 'queue', and 'user-tag'.

    // These are implemented using the existing PN-Set CRDT, by creating a new set every time we clear.
    // The list of PN-Set is indexed by a counter.
    _set_key(cat) {
      var coherent, index, index_key;
      index_key = this._key(cat);
      this.interface.setup_counter(index_key, this._expiry());
      [coherent, index] = this.interface.get_counter(index_key);
      if (index == null) {
        index = 0;
      }
      return this._key([cat, index].join('/'));
    }

    clear_(cat) {
      var index_key;
      index_key = this._key(cat);
      this.interface.setup_counter(index_key, this._expiry());
      this.interface.increment(index_key, 1);
    }

    add_(cat, value) {
      var key;
      key = this._set_key(cat);
      this.interface.setup_set(key, this._expiry());
      this.interface.add(key, value);
    }

    del_(cat, value) {
      var key;
      key = this._set_key(cat);
      this.interface.setup_set(key, this._expiry());
      this.interface.remove(key, value);
    }

    has_(cat, value) {
      var coherent, key;
      key = this._set_key(cat);
      [coherent, value] = this.interface.has(key, value);
      return value;
    }

    all_(cat) {
      var coherent, key, value;
      key = this._set_key(cat);
      [coherent, value] = this.interface.value(key);
      return value != null ? value : [];
    }

  };

  module.exports = {Reference, RedisInterfaceReference, BlueRingReference};

}).call(this);
